"""Join a sequence of arrays along a new axis."""
import numpy
import numpoly

from .common import implements


@implements(numpy.stack)
def stack(arrays, axis=0, out=None):
    """
    Join a sequence of arrays along a new axis.

    The ``axis`` parameter specifies the index of the new axis in the
    dimensions of the result. For example, if ``axis=0`` it will be the first
    dimension and if ``axis=-1`` it will be the last dimension.

    Args:
        arrays (Sequence[numpoly.ndpoly]):
            Each array must have the same shape.
        axis (Optional[int]):
            The axis in the result array along which the input arrays are
            stacked.
        out (Optional[numpy.ndarray]):
            If provided, the destination to place the result. The shape must be
            correct, matching that of what stack would have returned if no out
            argument were specified.

    Returns:
        (numpoly.ndpoly):
            The stacked array has one more dimension than the input arrays.

    Examples:
        >>> a = numpoly.symbols("x y z")
        >>> b = numpoly.polynomial([1, 2, 3])
        >>> numpoly.stack([a, b])
        polynomial([[x, y, z],
                    [1, 2, 3]])
        >>> numpoly.stack([a, b], axis=-1)
        polynomial([[x, 1],
                    [y, 2],
                    [z, 3]])

    """
    arrays = numpoly.align_exponents(*arrays)
    arrays = numpoly.align_dtype(*arrays)
    result = numpy.stack(
        [array.values for array in arrays], axis=axis, out=out)
    return numpoly.aspolynomial(result, names=arrays[0].indeterminants)




"""Compute the arithmetic mean along the specified axis."""
import numpy
import numpoly

from .common import implements, simple_dispatch


@implements(numpy.mean)
def mean(a, axis=None, dtype=None, out=None, **kwargs):
    """
    Compute the arithmetic mean along the specified axis.

    Returns the average of the array elements.  The average is taken over
    the flattened array by default, otherwise over the specified axis.
    `float64` intermediate and return values are used for integer inputs.

    Args:
        a (numpoly.ndpoly):
            Array containing numbers whose mean is desired. If `a` is not an
            array, a conversion is attempted.
        axis (Optional[numpy.ndarray]):
            Axis or axes along which the means are computed. The default is to
            compute the mean of the flattened array. If this is a tuple of
            ints, a mean is performed over multiple axes, instead of a single
            axis or all the axes as before.
        dtype (Optional[numpy.dtype]):
            Type to use in computing the mean.  For integer inputs, the default
            is `float64`; for floating point inputs, it is the same as the
            input dtype.
        out (Optional[numpy.ndpoly]):
            Alternate output array in which to place the result.  The default
            is ``None``; if provided, it must have the same shape as the
            expected output, but the type will be cast if necessary.
        keepdims (Optional[bool]):
            If this is set to True, the axes which are reduced are left in the
            result as dimensions with size one. With this option, the result
            will broadcast correctly against the input array.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (numpy.ndpoly):
            If `out=None`, returns a new array containing the mean values,
            otherwise a reference to the output array is returned.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> a = numpoly.polynomial([[1, 2*x], [3*y+x, 4]])
        >>> numpoly.mean(a)
        polynomial(1.25+0.75*x+0.75*y)
        >>> numpoly.mean(a, axis=0)
        polynomial([0.5+0.5*x+1.5*y, 2.0+x])
        >>> numpoly.mean(a, axis=1)
        polynomial([0.5+x, 2.0+0.5*x+1.5*y])

    """
    return simple_dispatch(
        numpy_func=numpy.mean,
        inputs=(a,),
        out=out,
        axis=axis,
        dtype=dtype,
        **kwargs
    )



"""First array elements raised to powers from second array, element-wise."""
import numpy
import numpoly

from .common import implements


@implements(numpy.power)
def power(x1, x2, **kwargs):
    """
    First array elements raised to powers from second array, element-wise.

    Raise each base in `x1` to the positionally-corresponding power in
    `x2`.  `x1` and `x2` must be broadcastable to the same shape. Note that an
    integer type raised to a negative integer power will raise a ValueError.

    Args:
        x1 (numpoly.ndpoly):
            The bases.
        x2 (numpoly.ndpoly):
            The exponents. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (numpoly.ndpoly):
            The bases in `x1` raised to the exponents in `x2`. This is a scalar
            if both `x1` and `x2` are scalars.

    Examples:
        >>> x = numpoly.symbols("x")
        >>> (x+1)**3
        polynomial(1+3*x+3*x**2+x**3)
        >>> poly = numpoly.symbols("x y")
        >>> (poly+1)**[1, 2]
        polynomial([1+x, 1+2*y+y**2])

    """
    x1 = numpoly.aspolynomial(x1)
    x2 = numpoly.aspolynomial(x2).tonumpy().astype(int)

    if not x2.shape:
        out = numpoly.ndpoly.from_attributes(
            [(0,)], [numpy.ones(x1.shape, dtype=x1._dtype)], x1.names[:1])
        for _ in range(x2.item()):
            out = numpoly.multiply(out, x1, **kwargs)

    elif x1.shape:
        if x2.shape[-1] == 1:
            if x1.shape[-1] == 1:
                out = numpoly.power(x1.T[0].T, x2.T[0].T).T[numpy.newaxis].T
            else:
                out = numpoly.concatenate([power(x, x2.T[0])[numpy.newaxis]
                                           for x in x1.T], axis=0).T
        elif x1.shape[-1] == 1:
            out = numpoly.concatenate([power(x1.T[0].T, x.T).T[numpy.newaxis]
                                       for x in x2.T], axis=0).T
        else:
            out = numpoly.concatenate([power(x1_, x2_).T[numpy.newaxis]
                                       for x1_, x2_ in zip(x1.T, x2.T)], axis=0).T
    else:
        out = numpoly.concatenate([power(x1, x.T).T[numpy.newaxis]
                                   for x in x2.T], axis=0).T
    return out



"""Convert inputs to arrays with at least one dimension."""
import numpy
import numpoly

from .common import implements


@implements(numpy.atleast_1d)
def atleast_1d(*arys):
    """
    Convert inputs to arrays with at least one dimension.

    Scalar inputs are converted to 1-dimensional arrays, whilst
    higher-dimensional inputs are preserved.

    Args:
        arys (numpoly.ndpoly):
            One or more input arrays.

    Returns:
        (ndarray):
            An array, or list of arrays, each with ``a.ndim >= 1``. Copies are
            made only if necessary.

    Examples:
        >>> numpoly.atleast_1d(numpoly.symbols("x"))
        polynomial([x])
        >>> numpoly.atleast_1d(1, [2, 3])
        [polynomial([1]), polynomial([2, 3])]

    """
    if len(arys) == 1:
        poly = numpoly.aspolynomial(arys[0])
        array = numpy.atleast_1d(poly.values)
        return numpoly.aspolynomial(array, names=poly.indeterminants)
    return [atleast_1d(ary) for ary in arys]



"""Evaluate polynomial."""
import logging
import numpy
import numpoly


def call(poly, *args, **kwargs):
    """
    Evaluate polynomial by inserting new values in to the indeterminants.

    Equaivalent to calling the polynomial or using the ``__call__`` method.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to evaluate.
        args (int, float, numpy.ndarray, numpoly.ndpoly):
            Argument to evaluate indeterminants. Ordered positional by
            ``poly.indeterminants``.
        kwargs (int, float, numpy.ndarray, numpoly.ndpoly):
            Same as ``args``, but positioned by name.

    Returns:
        (numpoly.ndpoly):
            Evaluated polynomial.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly = numpoly.polynomial([[x, x-1], [y, y+x]])
        >>> poly()
        polynomial([[x, -1+x],
                    [y, x+y]])
        >>> poly
        polynomial([[x, -1+x],
                    [y, x+y]])
        >>> poly(1, 0)
        array([[1, 0],
               [0, 1]])
        >>> poly(1, y=[0, 1, 2])
        array([[[1, 1, 1],
                [0, 0, 0]],
        <BLANKLINE>
               [[0, 1, 2],
                [1, 2, 3]]])
        >>> poly(y)
        polynomial([[y, -1+y],
                    [y, 2*y]])
        >>> poly(y=x-1, x=2*y)
        polynomial([[2*y, -1+2*y],
                    [-1+x, -1+2*y+x]])

    """
    logger = logging.getLogger(__name__)

    # Make sure kwargs contains all args and nothing but indeterminants:
    for arg, indeterminant in zip(args, poly.names):
        if indeterminant in kwargs:
            raise TypeError(
                "multiple values for argument '%s'" % indeterminant)
        kwargs[indeterminant] = arg
    extra_args = [key for key in kwargs if key not in poly.names]
    if extra_args:
        raise TypeError("unexpected keyword argument '%s'" % extra_args[0])

    if not kwargs:
        return poly.copy()

    # Saturate kwargs with values not given:
    indeterminants = poly.indeterminants
    for indeterminant in indeterminants:
        name = indeterminant.names[0]
        if name not in kwargs:
            kwargs[name] = indeterminant

    # There can only be one shape:
    ones = numpy.ones((), dtype=int)
    for value in kwargs.values():
        ones = ones * numpy.ones(numpoly.polynomial(value).shape, dtype=int)
    shape = poly.shape+ones.shape

    logger.debug("poly shape: %s", poly.shape)
    logger.debug("kwargs common shape: %s", ones.shape)
    logger.debug("output shape: %s", shape)

    # main loop:
    out = 0
    for exponent, coefficient in zip(poly.exponents, poly.coefficients):
        term = ones
        for power, name in zip(exponent, poly.names):
            term = term*kwargs[name]**power
        if isinstance(term, numpoly.ndpoly):
            tmp = numpoly.outer(coefficient, term)
        else:
            tmp = numpy.outer(coefficient, term)
        out = out+tmp.reshape(shape)

    if isinstance(out, numpoly.ndpoly):
        if out.isconstant():
            return out.tonumpy()
        out, _ = numpoly.align_indeterminants(out, indeterminants)
        return out
    return out



# pylint: disable=wildcard-import
"""Numpoly -- Multivariate polynomials as numpy elements."""
import logging
import os
import pkg_resources

from .baseclass import ndpoly

from .align import (
    align_polynomials,
    align_exponents,
    align_indeterminants,
    align_shape,
    align_dtype,
)
from .construct import (
    polynomial,
    aspolynomial,
    clean_attributes,
)
from .sympy_ import to_sympy

from .array_function import *
from .poly_function import *

try:
    __version__ = pkg_resources.get_distribution("numpoly").version
except pkg_resources.DistributionNotFound:
    __version__ = None


def configure_logging():
    """Configure logging for Numpoly."""
    logpath = os.environ.get("NUMPOLY_LOGPATH", os.devnull)
    logging.basicConfig(level=logging.DEBUG, filename=logpath, filemode="w")
    streamer = logging.StreamHandler()
    loglevel = logging.DEBUG if os.environ.get("NUMPOLY_DEBUG", "") else logging.WARNING
    streamer.setLevel(loglevel)

    logger = logging.getLogger(__name__)
    logger.addHandler(streamer)

configure_logging()



"""Subtract arguments, element-wise."""
import numpy
import numpoly

from .common import implements, simple_dispatch


@implements(numpy.subtract)
def subtract(x1, x2, out=None, where=True, **kwargs):
    """
    Subtract arguments, element-wise.

    Args:
        x1, x2 (numpoly.ndpoly):
            The arrays to be subtracted from each other. If
            ``x1.shape != x2.shape``, they must be broadcastable to a common
            shape (which becomes the shape of the output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Union[bool, numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (numpoly.ndpoly):
            The difference of `x1` and `x2`, element-wise. This is a scalar if
            both `x1` and `x2` are scalars.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> numpoly.subtract(x, 4)
        polynomial(x-4)
        >>> poly1 = y**numpy.arange(3)
        >>> poly2 = x**numpy.arange(9).reshape((3, 3))
        >>> numpoly.subtract(poly1, poly2)
        polynomial([[0, y-x, y**2-x**2],
                    [1-x**3, y-x**4, y**2-x**5],
                    [1-x**6, y-x**7, y**2-x**8]])

    """
    return simple_dispatch(
        numpy_func=numpy.subtract,
        inputs=(x1, x2),
        out=out,
        where=where,
        **kwargs
    )



"""Multi indices for monomial exponents."""
import numpy

from .cross_truncation import cross_truncate


def bindex(start, stop=None, dimensions=1, ordering="G", cross_truncation=1.):
    """
    Generate multi-indices for the monomial exponents.

    Args:
        start (Union[int, numpy.ndarray]):
            The lower order of the indices. If array of int, counts as lower
            bound for each axis.
        stop (Union[int, numpy.ndarray, None]):
            The maximum shape included. If omitted: stop <- start; start <- 0
            If int is provided, set as largest total order. If array of int,
            set as upper bound for each axis.
        dimensions (int):
            The number of dimensions in the expansion
        ordering (str):
            Criteria to sort the indices by.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion.

    Returns:
        list:
            Order list of indices.

    Examples:
        >>> bindex(4).tolist()
        [[0], [1], [2], [3]]
        >>> bindex(2, dimensions=2).tolist()
        [[0, 0], [0, 1], [1, 0]]
        >>> bindex(start=2, stop=3, dimensions=2).tolist()
        [[0, 2], [1, 1], [2, 0]]
        >>> bindex(start=0, stop=2, dimensions=3).tolist()
        [[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0]]

    """
    if stop is None:
        start, stop = 0, start
    start = numpy.array(start, dtype=int).flatten()
    stop = numpy.array(stop, dtype=int).flatten()
    ordering = ordering.upper()
    start[start < 0] = 0

    cross_truncation = cross_truncation*numpy.ones(2)
    indices = _bindex(start, stop, dimensions, cross_truncation)

    if "G" in ordering:
        indices = indices[numpy.lexsort([numpy.sum(indices, -1)])]

    if "I" in ordering:
        indices = indices[::-1]

    if "R" in ordering:
        indices = indices[:, ::-1]

    return indices


def _bindex(start, stop, dimensions=1, cross_truncation=1.):
    """Backend for the bindex function."""
    # At the beginning the current list of indices just ranges over the
    # last dimension.
    bound = stop.max()
    dtype = numpy.uint8 if bound < 256 else numpy.uint16
    range_ = numpy.arange(bound, dtype=dtype)
    indices = range_[:, numpy.newaxis]

    for _ in range(dimensions-1):

        # Repeats the current set of indices.
        # e.g. [0,1,2] -> [0,1,2,0,1,2,...,0,1,2]
        indices = numpy.tile(indices, (bound, 1))

        # Stretches ranges over the new dimension.
        # e.g. [0,1,2] -> [0,0,...,0,1,1,...,1,2,2,...,2]
        front = range_.repeat(len(indices)//bound)[:, numpy.newaxis]

        # Puts them two together.
        indices = numpy.column_stack((front, indices))

        # Truncate at each step to keep memory usage low
        lower = cross_truncate(indices, start-1, cross_truncation[0])
        upper = cross_truncate(indices, stop-1, cross_truncation[1])
        indices = indices[lower^upper]

    return indices.reshape(-1, dimensions)



"""Check if a polynomial is constant or not."""
import numpy
import numpoly


def isconstant(poly):
    """
    Check if a polynomial is constant or not.

    Args:
        poly (numpoly.ndpoly):
            polynomial to check if is constant or not.

    Returns:
        (bool):
            True if polynomial has no indeterminants.

    Examples:
        >>> x = numpoly.symbols("x")
        >>> numpoly.isconstant(numpoly.polynomial([x]))
        False
        >>> numpoly.isconstant(numpoly.polynomial([1]))
        True

    """
    poly = numpoly.aspolynomial(poly)
    for exponent, coefficient in zip(poly.exponents, poly.coefficients):
        if not numpy.any(exponent):
            continue
        if numpy.any(coefficient):
            return False
    return True



"""Add arguments element-wise."""
import numpy
import numpoly

from .common import implements, simple_dispatch


@implements(numpy.add)
def add(x1, x2, out=None, where=True, **kwargs):
    """
    Add arguments element-wise.

    Args:
        x1, x2 (numpoly.ndpoly):
            The arrays to be added. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Union[bool, numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (numpoly.ndpoly):
            The sum of `x1` and `x2`, element-wise. This is a scalar if both
            `x1` and `x2` are scalars.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> numpoly.add(x, 4)
        polynomial(x+4)
        >>> poly1 = x**numpy.arange(9).reshape((3, 3))
        >>> poly2 = y**numpy.arange(3)
        >>> numpoly.add(poly1, poly2)
        polynomial([[2, x+y, x**2+y**2],
                    [1+x**3, x**4+y, x**5+y**2],
                    [1+x**6, x**7+y, x**8+y**2]])

    """
    return simple_dispatch(
        numpy_func=numpy.add,
        inputs=(x1, x2),
        out=out,
        where=where,
        **kwargs
    )



"""Polynomial base class."""
from __future__ import division
import re
from six import string_types

import numpy

from . import align, construct, array_function, poly_function


REDUCE_MAPPINGS = {
    numpy.add: numpy.sum,
    numpy.multiply: numpy.prod,
    numpy.logical_and: numpy.all,
    numpy.logical_or: numpy.any,
}
ACCUMULATE_MAPPINGS = {
    numpy.add: numpy.cumsum,
    numpy.multiply: numpy.cumprod,
}
INDETERMINANT_DEFAULTS = {
    # Polynomial indeterminant defaults, if not defined.
    "base_name": "q",

    # Add a postfix index to single indeterminant name.
    # If single indeterminant name, e.g. 'q' is provided, but the polynomial is
    # multivariate, an extra postfix index is added to differentiate the names:
    # 'q0, q1, q2, ...'. If true, enforce this behavior for single variables as
    # well such that 'q' always get converted to 'q0'.
    "force_suffix": False,

    # Regular expression defining valid indeterminant names.
    "filter_regex": r"[\w_]",
}


class ndpoly(numpy.ndarray):  # pylint: disable=invalid-name
    """
    Polynomial as numpy array.

    An array object represents a multidimensional, homogeneous polynomial array
    of fixed-size items. An associated data-type object describes the format of
    each element in the array (its byte-order, how many bytes it occupies in
    memory, whether it is an integer, a floating point number, or something
    else, etc.)

    Arrays should be constructed using `symbols`, `monomial`, `polynomial`,
    etc.

    Attributes:
        coefficients (List[numpy.ndarray, ...]):
            The polynomial coefficients. Together with exponents defines the
            polynomial form.
        exponents (numpy.ndarray):
            The polynomial exponents. 2-dimensional where the first axis is the
            same length as coefficients and the second is the length of the
            indeterminant names.
        keys (List[str, ...]):
            The raw names of the coefficients. One-to-one with `exponents`, but
            as string as to be compatible with numpy structured array. Unlike
            the exponents, that are useful for mathematical manipulation, the
            keys are useful as coefficient lookup.
        indeterminats (numpoly.ndpoly):
            Secondary polynomial only consisting of an array of simple
            independent variables found in the polynomial array.
        names (Tuple[str, ...]):
            Same as `indeterminants`, but only the names as string.
        values (numpy.ndarray):
            Expose the underlying structured array.

    Examples:
        >>> poly = ndpoly(
        ...     exponents=[(0, 1), (0, 0)], shape=(3,), names="x y")
        >>> poly[";;"] = 1, 2, 3
        >>> poly[";<"] = 4, 5, 6
        >>> numpy.array(poly.coefficients)
        array([[4, 5, 6],
               [1, 2, 3]])
        >>> poly
        polynomial([4*y+1, 5*y+2, 6*y+3])
        >>> poly[0]
        polynomial(4*y+1)

    """

    # =================================================
    # Stuff to get subclassing of ndarray to run smooth
    # =================================================

    __array_priority__ = 16
    _dtype = None
    keys = None
    names = None

    # Numpy structured array names doesn't like characters reserved by Python.
    # The largest index found with this property is 58: ':'.
    # Above this, everything looks like it works as expected.
    KEY_OFFSET = 59

    def __new__(
            cls,
            exponents=((0,),),
            shape=(),
            names=None,
            dtype=None,
            **kwargs
    ):
        """
        Class constructor.

        Args:
            exponents (numpy.ndarray):
                The exponents in an integer array with shape ``(N, D)``, where
                ``N`` is the number of terms in the polynomial sum and ``D`` is
                the number of dimensions.
            shape (Tuple[int, ...]):
                Shape of created array.
            names (Union[None, str, Tuple[str], numpoly.ndpoly]):
                The name of the indeterminant variables in te polynomial. If
                polynomial, inherent from it. Else, pass argument to
                `numpoly.symbols` to create the indeterminants names. If only
                one name is provided, but more than one is required,
                indeterminants will be extended with an integer index. If
                omitted, use ``INDETERMINANT_DEFAULTS["base_name"]``.
            dtype:
                Any object that can be interpreted as a numpy data type.
            kwargs:
                Extra arguments passed to `numpy.ndarray` constructor.

        """
        exponents = numpy.array(exponents, dtype=numpy.uint32)

        if names is None:
            names = INDETERMINANT_DEFAULTS["base_name"]
        if isinstance(names, string_types):
            names = poly_function.symbols(names)
        if isinstance(names, ndpoly):
            names = names.names
        if len(names) == 1 and (INDETERMINANT_DEFAULTS["force_suffix"] or
                                exponents.shape[1] > 1):
            names = tuple("%s%d" % (str(names[0]), idx)
                          for idx in range(exponents.shape[1]))
        for name in names:
            assert re.search(INDETERMINANT_DEFAULTS["filter_regex"], name), (
                "invalid polynomial name; "
                "expected format: %r" % INDETERMINANT_DEFAULTS["filter_regex"])

        if numpy.prod(exponents.shape):
            keys = (exponents+cls.KEY_OFFSET).flatten()
            keys = keys.view("U%d" % exponents.shape[-1])
        else:
            keys = numpy.zeros(0, dtype="U1")

        dtype = int if dtype is None else dtype
        dtype_ = numpy.dtype([(key, dtype) for key in keys])

        obj = super(ndpoly, cls).__new__(
            cls, shape=shape, dtype=dtype_, **kwargs)
        obj._dtype = numpy.dtype(dtype)  # pylint: disable=protected-access
        obj.keys = keys
        obj.names = tuple(names)
        return obj

    def __array_finalize__(self, obj):
        """Finalize numpy constructor."""
        if obj is None:
            return
        self.keys = getattr(obj, "keys", None)
        self.names = getattr(obj, "names", None)
        self._dtype = getattr(obj, "_dtype", None)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Dispatch method for operators."""
        if method == "reduce":
            ufunc = REDUCE_MAPPINGS[ufunc]
        elif method == "accumulate":
            ufunc = ACCUMULATE_MAPPINGS[ufunc]
        else:
            assert method == "__call__", (
                "method %s not recognised" % method)
        assert ufunc in array_function.ARRAY_FUNCTIONS, (
            "function %s not supported by numpoly." % ufunc)
        return array_function.ARRAY_FUNCTIONS[ufunc](*inputs, **kwargs)

    def __array_function__(self, func, types, args, kwargs):
        """Dispatch method for functions."""
        assert func in array_function.ARRAY_FUNCTIONS, (
            "function %s not supported by numpoly." % func)
        return array_function.ARRAY_FUNCTIONS[func](*args, **kwargs)

    # ======================================
    # Properties specific for ndpoly objects
    # ======================================

    @property
    def coefficients(self):
        """
        Polynomial coefficients.

        Examples:
            >>> x, y = numpoly.symbols("x y")
            >>> poly = numpoly.polynomial([2*x**4, -3*y**2+14])
            >>> poly
            polynomial([2*x**4, 14-3*y**2])
            >>> numpy.array(poly.coefficients)
            array([[ 0, 14],
                   [ 0, -3],
                   [ 2,  0]])

        """
        if not self.size:
            return []
        out = numpy.empty((len(self.keys),) + self.shape, dtype=self._dtype)
        for idx, key in enumerate(self.keys):
            out[idx] = numpy.ndarray.__getitem__(self, key)
        return list(out)

    @property
    def exponents(self):
        """
        Polynomial exponents.

        Examples:
            >>> x, y = numpoly.symbols("x y")
            >>> poly = numpoly.polynomial([2*x**4, -3*y**2+14])
            >>> poly
            polynomial([2*x**4, 14-3*y**2])
            >>> poly.exponents
            array([[0, 0],
                   [0, 2],
                   [4, 0]], dtype=uint32)

        """
        exponents = self.keys.flatten().view(numpy.uint32)-self.KEY_OFFSET
        if numpy.prod(exponents.shape):
            exponents = exponents.reshape(len(self.keys), -1)
        elif self.size == 1:
            exponents = exponents.reshape(1, 1)
        else:
            exponents = numpy.zeros((1, 1), dtype=numpy.uint32)
        assert len(exponents) > 0
        assert len(exponents.shape) == 2
        return exponents

    @staticmethod
    def from_attributes(
            exponents,
            coefficients,
            names="q",
            dtype=None,
            clean=True,
    ):
        """
        Construct polynomial from polynomial attributes.

        Args:
            exponents (numpy.ndarray):
                The exponents in an integer array with shape ``(N, D)``, where
                ``N`` is the number of terms in the polynomial sum and ``D`` is
                the number of dimensions.
            coefficients (Iterable[numpy.ndarray]):
                The polynomial coefficients. Must correspond to `exponents` by
                having the same length ``N``.
            names (Union[str, Tuple[str, ...], numpoly.ndpoly]):
                The indeterminants names, either as string names or as
                simple polynomials. Must correspond to the exponents by having
                length ``D``.
            dtype (Optional[numpy.dtype]):
                The data type of the polynomial. If omitted, extract from
                `coefficients`.
            clean (bool):
                Clean up attributes, removing redundant names and exponents.
                Used to ensure alignment isn't broken.

        Returns:
            (numpoly.ndpoly):
                Polynomials with attributes defined by input.

        Examples:
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[0]], coefficients=[4])
            polynomial(4)
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[1]], coefficients=[[1, 2, 3]])
            polynomial([q, 2*q, 3*q])
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[0], [1]], coefficients=[[0, 1], [1, 1]])
            polynomial([q, 1+q])
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[0, 1], [1, 1]], coefficients=[[0, 1], [1, 1]])
            polynomial([q0*q1, q1+q0*q1])

        """
        return construct.polynomial_from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            names=names,
            dtype=dtype,
            clean=clean,
        )

    @property
    def indeterminants(self):
        """
        Polynomial indeterminants.

        Examples:
            >>> x, y = numpoly.symbols("x y")
            >>> poly = numpoly.polynomial([2*x**4, -3*y**2+14])
            >>> poly
            polynomial([2*x**4, 14-3*y**2])
            >>> poly.indeterminants
            polynomial([x, y])

        """
        return construct.polynomial_from_attributes(
            exponents=numpy.eye(len(self.names), dtype=int),
            coefficients=numpy.eye(len(self.names), dtype=int),
            names=self.names,
        )

    @property
    def values(self):
        """
        Expose the underlying structured array.

        Typically used for operator dispatching and not for use to conversion.

        Examples:
            >>> numpoly.symbols("x").values
            array((1,), dtype=[('<', '<i8')])
            >>> numpoly.symbols("x y").values
            array([(1, 0), (0, 1)], dtype=[('<;', '<i8'), (';<', '<i8')])

        """
        return numpy.ndarray(
            shape=self.shape,
            dtype=[(key, self.dtype) for key in self.keys],
            buffer=self.data
        )

    def isconstant(self):
        """
        Check if a polynomial is constant or not.

        Returns:
            (bool):
                True if all elements in array are constant.

        Examples:
            >>> x = numpoly.symbols("x")
            >>> x.isconstant()
            False
            >>> numpoly.polynomial([1, 2]).isconstant()
            True
            >>> numpoly.polynomial([1, x]).isconstant()
            False

        """
        return poly_function.isconstant(self)

    def todict(self):
        """
        Cast to dict where keys are exponents and values are coefficients.

        Returns:
            (Dict[Tuple[int, ...], numpy.ndarray]):
                Dictionary where keys are exponents and values are
                coefficients.

        Examples:
            >>> x, y = numpoly.symbols("x y")
            >>> poly = 2*x**4-3*y**2+14
            >>> poly
            polynomial(14+2*x**4-3*y**2)
            >>> poly.todict() == {(0, 0): 14, (4, 0): 2, (0, 2): -3}
            True

        """
        return {tuple(exponent): coefficient
                for exponent, coefficient in zip(
                    self.exponents, self.coefficients)}

    def tonumpy(self):
        """
        Cast polynomial to numpy.ndarray, if possible.

        Raises:
            ValueError:
                When polynomial include indeterminats, casting to numpy.

        Returns:
            (numpy.ndarray):
                Same as object, but cast to `numpy.ndarray`.

        Examples:
            >>> numpoly.polynomial([1, 2]).tonumpy()
            array([1, 2])
            >>> numpoly.symbols("x").tonumpy()
            Traceback (most recent call last):
                ...
            ValueError: only constant polynomials can be converted to array.

        """
        return poly_function.tonumpy(self)

    # =============================================
    # Override numpy properties to work with ndpoly
    # =============================================

    @property
    def dtype(self):
        """Show coefficient dtype instead of the structured array."""
        return self._dtype

    def astype(self, dtype, **kwargs):
        """Wrap ndarray.astype."""
        coefficients = [coefficient.astype(dtype, **kwargs)
                        for coefficient in self.coefficients]
        return construct.polynomial_from_attributes(
            self.exponents, coefficients, self.names, dtype=dtype)

    def round(self, decimals=0, out=None):
        """Wrap ndarray.round."""
        # Not sure why it is required. Likely a numpy bug.
        return array_function.around(self, decimals=decimals, out=out)

    # ============================================================
    # Override dunder methods that isn't dealt with by dispatching
    # ============================================================

    def __call__(self, *args, **kwargs):
        """Evaluate polynomial."""
        return poly_function.call(self, *args, **kwargs)

    def __eq__(self, other):
        """Left equality."""
        return array_function.equal(self, other)

    def __getitem__(self, index):
        """
        Get array item or slice.

        Args:
            index (Union[int, str, Tuple[int, ...], numpy.ndarray]):
                The index to extract. If string type, extract using th
                underlying structured array data type. Else, extract as
                numpy.ndarray.

        Examples:
            >>> x, y = numpoly.symbols("x y")
            >>> poly = numpoly.polynomial([[1-4*x, x**2], [y-3, x*y*y]])
            >>> poly
            polynomial([[1-4*x, x**2],
                        [-3+y, x*y**2]])
            >>> poly[0]
            polynomial([1-4*x, x**2])
            >>> poly[:, 1]
            polynomial([x**2, x*y**2])
            >>> poly["<;"]
            array([[-4,  0],
                   [ 0,  0]])

        """
        if isinstance(index, string_types):
            return numpy.asarray(super(ndpoly, self).__getitem__(index))
        return construct.polynomial_from_attributes(
            exponents=self.exponents,
            coefficients=[coeff[index] for coeff in self.coefficients],
            names=self.names,
        )

    def __iter__(self):
        """Iterate polynomial array."""
        coefficients = numpy.array(list(self.coefficients))
        return iter(construct.polynomial_from_attributes(
            exponents=self.exponents,
            coefficients=coefficients[:, idx],
            names=self.names,
        ) for idx in range(len(self)))

    def __ne__(self, other):
        """Not equal."""
        return array_function.not_equal(self, other)

    def __repr__(self):
        """Canonical string representation."""
        return array_function.array_repr(self)

    def __str__(self):
        """Pretty string representation."""
        return array_function.array_str(self)



"""Multiply arguments element-wise."""
import numpy
import numpoly

from .common import implements


@implements(numpy.multiply)
def multiply(x1, x2, out=None, where=True, **kwargs):
    """
    Multiply arguments element-wise.

    Args:
        x1, x2 (numpoly.ndpoly):
            Input arrays to be multiplied. If ``x1.shape != x2.shape``, they
            must be broadcastable to a common shape (which becomes the shape of
            the output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (numpoly.ndpoly):
            The product of `x1` and `x2`, element-wise. This is a scalar if
            both `x1` and `x2` are scalars.

    Examples:
        >>> x1 = numpy.arange(9.0).reshape((3, 3))
        >>> xyz = numpoly.symbols("x y z")
        >>> numpoly.multiply(x1, xyz)
        polynomial([[0.0, y, 2.0*z],
                    [3.0*x, 4.0*y, 5.0*z],
                    [6.0*x, 7.0*y, 8.0*z]])

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    no_output = out is None
    if no_output:
        exponents = (numpy.tile(x1.exponents, (len(x2.exponents), 1))+
                     numpy.repeat(x2.exponents, len(x1.exponents), 0))
        out = numpoly.ndpoly(
            exponents=numpy.unique(exponents, axis=0),
            shape=x1.shape,
            names=x1.indeterminants,
            dtype=x1.dtype,
        )

    seen = set()
    for expon1, coeff1 in zip(x1.exponents, x1.coefficients):
        for expon2, coeff2 in zip(x2.exponents, x2.coefficients):
            key = (expon1+expon2+x1.KEY_OFFSET).flatten()
            key = key.view("U%d" % len(expon1)).item()
            if key in seen:
                out[key] += numpy.multiply(
                    coeff1, coeff2, where=where, **kwargs)
            else:
                numpy.multiply(
                    coeff1, coeff2, out=out[key], where=where, **kwargs)
            seen.add(key)
    if no_output:
        out = numpoly.clean_attributes(out)
    return out



"""Stack arrays in sequence depth wise (along third axis)."""
import numpy
import numpoly

from .common import implements


@implements(numpy.dstack)
def dstack(tup):
    """
    Stack arrays in sequence depth wise (along third axis).

    This is equivalent to concatenation along the third axis after 2-D arrays
    of shape `(M,N)` have been reshaped to `(M,N,1)` and 1-D arrays of shape
    `(N,)` have been reshaped to `(1,N,1)`. Rebuilds arrays divided by
    `dsplit`.

    This function makes most sense for arrays with up to 3 dimensions. For
    instance, for pixel-data with a height (first axis), width (second axis),
    and r/g/b channels (third axis). The functions `concatenate`, `stack` and
    `block` provide more general stacking and concatenation operations.

    Args:
        tup (Sequence[numpoly.ndpoly]):
            The arrays must have the same shape along all but the third axis.
            1-D or 2-D arrays must have the same shape.

    Returns:
        (numpoly.ndpoly):
            The array formed by stacking the given arrays, will be at least
            3-D.

    Examples:
        >>> a = numpoly.symbols("x y z")
        >>> b = numpoly.polynomial([1, 2, 3])
        >>> numpoly.dstack([a, b])
        polynomial([[[x, 1],
                     [y, 2],
                     [z, 3]]])
        >>> c = numpoly.polynomial([[1], [2], [3]])
        >>> d = a.reshape(3, 1)
        >>> numpoly.dstack([c, d])
        polynomial([[[1, x]],
        <BLANKLINE>
                    [[2, y]],
        <BLANKLINE>
                    [[3, z]]])

    """
    arrays = numpoly.align_exponents(*tup)
    arrays = numpoly.align_dtype(*arrays)
    result = numpy.dstack([array.values for array in arrays])
    return numpoly.aspolynomial(result, names=arrays[0].indeterminants)



"""Return the string representation of an array."""
import numpy
import numpoly

from .common import implements


@implements(numpy.array_repr)
def array_repr(arr, max_line_width=None, precision=None, suppress_small=None):
    """
    Return the string representation of an array.

    Args:
        arr : ndarray
            Input array.
        max_line_width : int, optional
            Inserts newlines if text is longer than `max_line_width`. Defaults
            to ``numpy.get_printoptions()['linewidth']``.
        precision : int, optional
            Floating point precision. Defaults to
            ``numpy.get_printoptions()['precision']``.
        suppress_small : bool, optional
            Represent numbers "very close" to zero as zero; default is False.
            Very close is defined by precision: if the precision is 8, e.g.,
            numbers smaller (in absolute value) than 5e-9 are represented as
            zero. Defaults to ``numpy.get_printoptions()['suppress']``.

    Returns:
        (str):
            The string representation of an array.

    Examples:
        >>> x = numpoly.symbols("x")
        >>> numpoly.array_repr(numpoly.polynomial([1, x]))
        'polynomial([1, x])'
        >>> numpoly.array_repr(numpoly.polynomial([]))
        'polynomial([], dtype=int64)'
        >>> numpoly.array_repr(
        ...     numpoly.polynomial([1e-6, 4e-7*x, 2*x, 3]),
        ...     precision=4,
        ...     suppress_small=True,
        ... )
        'polynomial([0.0, 0.0, 2.0*x, 3.0])'

    """
    prefix = "polynomial("
    suffix = ")"
    arr = numpoly.aspolynomial(arr)
    if not arr.size:
        return prefix + "[], dtype=%s" % arr.dtype.name + suffix

    if precision is None:
        precision = numpy.get_printoptions()["precision"]
    if suppress_small is None:
        suppress_small = numpy.get_printoptions()["suppress"]
    arr = to_string(arr, precision=precision, suppress_small=suppress_small)

    return prefix + numpy.array2string(
        numpy.array(arr),
        max_line_width=max_line_width,
        separator=", ",
        formatter={"all": str},
        prefix=prefix,
        suffix=suffix,
    ) + suffix


def to_string(poly, precision=None, suppress_small=None):
    """
    Convert numpoly object to string object, or array of string objects.

    Args:
        poly (numpoly.ndpoly):
            Polynomial object to convert to strings.

    Returns:
        (numpy.ndarray, str):
            If scalar, a string, or if array, numpy.array with string values.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly = numpoly.polynomial([[1, x**3], [y-1, -3*x]])
        >>> string_array = to_string(poly)
        >>> string_array
        [['1', 'x**3'], ['-1+y', '-3*x']]
        >>> type(string_array[0][0]) == str
        True

    """
    if poly.shape:
        return [
            to_string(
                poly_, precision=precision, suppress_small=suppress_small)
            for poly_ in poly
        ]

    output = []
    for exponents, coefficient in zip(
            poly.exponents.tolist(), poly.coefficients):

        if not coefficient or (
                suppress_small and
                abs(coefficient) < 10**-precision  # pylint: disable=invalid-unary-operand-type
        ):
            continue

        if coefficient == 1 and any(exponents):
            out = ""
        elif coefficient == -1 and any(exponents):
            out = "-"
        else:
            out = str(coefficient)

        for exponent, indeterminant in zip(exponents, poly.names):
            if exponent:
                if out not in ("", "-"):
                    out += "*"
                out += indeterminant
            if exponent > 1:
                out += "**"+str(exponent)
        if output and float(coefficient) >= 0:
            out = "+"+out
        output.append(out)

    if output:
        return "".join(output)
    return str(numpy.zeros(1, dtype=poly.dtype).item())



"""Numerical positive, element-wise."""
import numpy

from .common import implements, simple_dispatch


@implements(numpy.positive)  # pylint: disable=no-member
def positive(x, out=None, where=True, **kwargs):
    """
    Numerical positive, element-wise.

    Args:
        x (numpoly.ndpoly):
            Input array.
        out (Optional[numpoly.ndpoly]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where : array_like, optional
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        y (numpoly.ndpoly):
            Returned array or scalar: `y = +x`. This is a scalar if `x` is
            a scalar.

    Examples:
        >>> x = numpoly.symbols("x")
        >>> numpoly.positive([-0, 0, -x, x])
        polynomial([0, 0, -x, x])

    """
    return simple_dispatch(
        numpy_func=numpy.positive,  # pylint: disable=no-member
        inputs=(x,),
        out=out,
        where=where,
        **kwargs
    )



"""View inputs as arrays with at least three dimensions."""
import numpy
import numpoly

from .common import implements


@implements(numpy.atleast_3d)
def atleast_3d(*arys):
    """
    View inputs as arrays with at least three dimensions.

    Args:
        arys (numpoly.ndpoly):
            One or more array-like sequences. Non-array inputs are converted
            to arrays. Arrays that already have three or more dimensions are
            preserved.

    Returns:
        (numpoly.ndpoly):
            An array, or list of arrays, each with ``a.ndim >= 3``.  Copies are
            avoided where possible, and views with three or more dimensions are
            returned.  For example, a 1-D array of shape ``(N,)`` becomes
            a view of shape ``(1, N, 1)``, and a 2-D array of shape ``(M, N)``
            becomes a view of shape ``(M, N, 1)``.

    Examples:
        >>> numpoly.atleast_3d(numpoly.symbols("x"))
        polynomial([[[x]]])
        >>> numpoly.atleast_3d(1, [2, 3])
        [polynomial([[[1]]]), polynomial([[[2],
                     [3]]])]

    """
    if len(arys) == 1:
        poly = numpoly.aspolynomial(arys[0])
        array = numpy.atleast_3d(poly.values)
        return numpoly.aspolynomial(array, names=poly.indeterminants)
    return [atleast_3d(ary) for ary in arys]



"""View inputs as arrays with at least two dimensions."""
import numpy
import numpoly

from .common import implements


@implements(numpy.atleast_2d)
def atleast_2d(*arys):
    """
    View inputs as arrays with at least two dimensions.

    Args:
        arys (numpoly.ndpoly):
            One or more array-like sequences. Non-array inputs are converted
            to arrays. Arrays that already have two or more dimensions are
            preserved.

    Returns:
        (numpoly.ndpoly):
            An array, or list of arrays, each with ``a.ndim >= 2``. Copies are
            avoided where possible, and views with two or more dimensions are
            returned.

    Examples:
        >>> numpoly.atleast_2d(numpoly.symbols("x"))
        polynomial([[x]])
        >>> numpoly.atleast_2d(1, [2, 3], [[4]])
        [polynomial([[1]]), polynomial([[2, 3]]), polynomial([[4]])]

    """
    if len(arys) == 1:
        poly = numpoly.aspolynomial(arys[0])
        array = numpy.atleast_2d(poly.values)
        return numpoly.aspolynomial(array, names=poly.indeterminants)
    return [atleast_2d(ary) for ary in arys]



"""Monomial construction."""
from __future__ import division

import numpy
import numpoly

from .bindex import bindex


def monomial(start, stop=None, ordering="G", cross_truncation=1., names=None):
    """
    Create an polynomial monomial expansion.

    Args:
        start (int, numpy.ndarray):
            The minimum polynomial to include. If int is provided, set as
            lowest total order. If array of int, set as lower order along each
            indeterminant.
        stop (int, numpy.ndarray):
            The maximum shape included. If omitted:
            ``stop <- start; start <- 0`` If int is provided, set as largest
            total order. If array of int, set as largest order along each
            indeterminant.
        ordering (str):
            The monomial ordering where the letters ``G``, ``I`` and ``R`` can
            be used to set grade, inverse and reverse to the ordering. For
            ``names=("x", "y"))`` we get for various values for
            ``ordering``:

            ========  =====================
            ordering  output
            ========  =====================
            ""        [1 y y**2 x x*y x**2]
            "G"       [1 y x y**2 x*y x**2]
            "I"       [x**2 x*y x y**2 y 1]
            "R"       [1 x x**2 y x*y y**2]
            "GIR"     [y**2 x*y x**2 y x 1]
            ========  =====================
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion.
        names (None, numpoly.ndpoly, str, Tuple[str, ...])
            The indeterminants names used to create the monomials expansion.

    Returns:
        (numpoly.ndpoly):
            Monomial expansion.

    Examples:
        >>> numpoly.monomial(4)
        polynomial([1, q, q**2, q**3])
        >>> numpoly.monomial(4, 5, ordering="GR", names=("x", "y"))
        polynomial([x**4, x**3*y, x**2*y**2, x*y**3, y**4])
        >>> numpoly.monomial(2, [3, 4])
        polynomial([q1**2, q0*q1, q0**2, q1**3])
        >>> numpoly.monomial(0, 5, names=("x", "y"), cross_truncation=0.5)
        polynomial([1, y, x, y**2, x*y, x**2, y**3, x**3, y**4, x**4])

    """
    if stop is None:
        start, stop = numpy.array(0), start

    start = numpy.array(start, dtype=int)
    stop = numpy.array(stop, dtype=int)
    dimensions = 1 if names is None else len(names)
    dimensions = max(start.size, stop.size, dimensions)

    indices = bindex(
        start=start,
        stop=stop,
        dimensions=dimensions,
        ordering=ordering,
        cross_truncation=cross_truncation,
    )

    poly = numpoly.ndpoly(
        exponents=indices,
        shape=(len(indices),),
        names=names,
    )
    for coeff, key in zip(
            numpy.eye(len(indices), dtype=int), poly.keys):
        poly[key] = coeff
    return poly



"""Functions for constructing polynomials."""
from .polynomial import polynomial
from .aspolynomial import aspolynomial
from .clean import clean_attributes
from .from_attributes import polynomial_from_attributes



"""Compose polynomial from array of arrays of polynomials."""
import numpy
import numpoly

from .clean import postprocess_attributes


def compose_polynomial_array(
        arrays,
        dtype=None,
):
    """
    Compose polynomial from array of arrays of polynomials.

    Backend for `numpoly.polynomial` when input is undetermined.

    Args:
        arrays (Any):
            Input to be converted to a `numpoly.ndpoly` polynomial type.
        dtype (Optional[numpy.dtype]):
            Data type used for the polynomial coefficients.

    Return:
        (numpoly.ndpoly):
            Polynomial based on input `arrays`.
    """
    arrays_ = numpy.array(arrays, dtype=object)
    shape = arrays_.shape
    if not arrays_.size:
        return numpoly.ndpoly(shape=shape, dtype=dtype)
    if len(arrays_.shape) > 1:
        return numpoly.concatenate([
            numpoly.aspolynomial(array, dtype)[numpy.newaxis]
            for array in arrays
        ], axis=0)

    arrays = arrays_.flatten()

    indices = numpy.array([isinstance(array, numpoly.ndpoly)
                           for array in arrays])
    arrays[indices] = numpoly.align_indeterminants(*arrays[indices])
    names = arrays[indices][0] if numpy.any(indices) else None
    arrays = arrays.tolist()

    dtypes = []
    keys = {(0,)}
    for array in arrays:
        if isinstance(array, numpoly.ndpoly):
            dtypes.append(array.dtype)
            keys = keys.union([tuple(key) for key in array.exponents.tolist()])
        elif isinstance(array, (numpy.generic, numpy.ndarray)):
            dtypes.append(array.dtype)
        else:
            dtypes.append(type(array))

    if dtype is None:
        dtype = numpy.find_common_type(dtypes, [])
    length = max([len(key) for key in keys])

    collection = {}
    for idx, array in enumerate(arrays):
        if isinstance(array, numpoly.ndpoly):
            for key, value in zip(array.exponents, array.coefficients):
                key = tuple(key)+(0,)*(length-len(key))
                if key not in collection:
                    collection[key] = numpy.zeros(len(arrays), dtype=dtype)
                collection[key][idx] = value
        else:
            key = (0,)*length
            if key not in collection:
                collection[key] = numpy.zeros(len(arrays), dtype=dtype)
            collection[key][idx] = array

    exponents = sorted(collection)
    coefficients = numpy.array([collection[key] for key in exponents])
    coefficients = coefficients.reshape(-1, *shape)

    exponents, coefficients, names = postprocess_attributes(
        exponents, coefficients, names)
    return numpoly.ndpoly.from_attributes(
        exponents=exponents,
        coefficients=coefficients,
        names=names,
    )



"""Return the largest integer smaller or equal to the division of the inputs."""
import numpy
import numpoly

from .common import implements


@implements(numpy.floor_divide)
def floor_divide(x1, x2, out=None, where=True, **kwargs):
    """
    Return the largest integer smaller or equal to the division of the inputs.

    It is equivalent to the Python ``//`` operator and pairs with the
    Python ``%`` (`remainder`), function so that ``a = a % b + b * (a // b)``
    up to roundoff.

    Args:
        x1 (numpoly.ndpoly):
            Numerator.
        x2 (numpoly.ndpoly):
            Denominator. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value.
            Note that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (numpoly.ndpoly):
            This is a scalar if both `x1` and `x2` are scalars.

    Examples:
        >>> xyz = [1, 2, 4]*numpoly.symbols("x y z")
        >>> numpoly.floor_divide(xyz, 2.)
        polynomial([0.0, y, 2.0*z])
        >>> numpoly.floor_divide(xyz, [1, 2, 4])
        polynomial([x, y, z])
        >>> numpoly.floor_divide([1, 2, 4], xyz)
        Traceback (most recent call last):
            ...
        ValueError: only constant polynomials can be converted to array.

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    x2 = x2.tonumpy()
    no_output = out is None
    if no_output:
        out = numpoly.ndpoly(
            exponents=x1.exponents,
            shape=x1.shape,
            names=x1.indeterminants,
            dtype=numpy.common_type(x1, numpy.array(1.)),
        )
    for key in x1.keys:
        numpy.floor_divide(x1[key], x2, out=out[key], where=where, **kwargs)
    if no_output:
        out = numpoly.clean_attributes(out)
    return out


# def polynomial_division(x1, x2):
#     """
#     Examples:
#         >>> # polynomial_division(numpoly.monomial(5, names="x"), numpoly.symbols("x"))
#     """
#     x1, x2 = numpoly.align_polynomials(x1, x2)
#     out = 0
#     for idx, name in enumerate(x1.names):
#         pass


#     exponents = x1.exponents.copy()
#     coefficients = x1.coefficients.copy()
#     orders1 = numpy.sum(x1.exponents, -1)
#     orders2 = numpy.sum(x2.exponents, -1)

#     while numpy.max(orders1) > numpy.max(orders2):
#         exponents



"""Return True if two arrays are element-wise equal within a tolerance."""
import numpy
import numpoly

from .common import implements


@implements(numpy.allclose)
def allclose(a, b, rtol=1e-5, atol=1e-8, equal_nan=False):
    """
    Return True if two arrays are element-wise equal within a tolerance.

    The tolerance values are positive, typically very small numbers.  The
    relative difference (`rtol` * abs(`b`)) and the absolute difference
    `atol` are added together to compare against the absolute difference
    between `a` and `b`.

    If either array contains one or more NaNs, False is returned.
    Infs are treated as equal if they are in the same place and of the same
    sign in both arrays.

    Args:
        a, b (numpoly.ndpoly):
            Input arrays to compare.
        rtol (float):
            The relative tolerance parameter (see Notes).
        atol : float
            The absolute tolerance parameter (see Notes).
        equal_nan : bool
            Whether to compare NaN's as equal.  If True, NaN's in `a` will be
            considered equal to NaN's in `b` in the output array.

    Returns:
        (bool):
            Returns True if the two arrays are equal within the given
            tolerance; False otherwise.

    Notes:
        If the following equation is element-wise True, then allclose returns
        True.

        absolute(`a` - `b`) <= (`atol` + `rtol` * absolute(`b`))

        The above equation is not symmetric in `a` and `b`, so that
        ``allclose(a, b)`` might be different from ``allclose(b, a)`` in some
        rare cases.

        The comparison of `a` and `b` uses standard broadcasting, which means
        that `a` and `b` need not have the same shape in order for
        ``allclose(a, b)`` to evaluate to True.  The same is true for `equal`
        but not `array_equal`.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> numpoly.allclose([1e10*x, 1e-7], [1.00001e10*x, 1e-8])
        False
        >>> numpoly.allclose([1e10*x, 1e-8], [1.00001e10*x, 1e-9])
        True
        >>> numpoly.allclose([1e10*x, 1e-8], [1.00001e10*y, 1e-9])
        False
        >>> numpoly.allclose([x, numpy.nan],
        ...                  [x, numpy.nan], equal_nan=True)
        True

    """
    a, b = numpoly.align_polynomials(a, b)
    for coeff1, coeff2 in zip(a.coefficients, b.coefficients):
        if not numpy.allclose(
                coeff1, coeff2, atol=atol, rtol=rtol, equal_nan=equal_nan):
            return False
    return True



"""Polynomials differentiation functions."""
from six import string_types

import numpy
import numpoly


def diff(poly, *diffvars):
    """
    Polynomial differential operator.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to differentiate.
        diffvars (numpoly.ndpoly, str):
            Singleton variables to take derivative off.

    Returns:
        Same as ``poly`` but differentiated with respect to ``diffvars``.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly = numpoly.polynomial([1, x, x*y**2+1])
        >>> poly
        polynomial([1, x, 1+x*y**2])
        >>> numpoly.diff(poly, "x")
        polynomial([0, 1, y**2])
        >>> numpoly.diff(poly, 0, 1)
        polynomial([0, 0, 2*y])
        >>> numpoly.diff(poly, x, x, x)
        polynomial([0, 0, 0])

    """
    poly = poly_ref = numpoly.aspolynomial(poly)

    for diffvar in diffvars:
        if isinstance(diffvar, string_types):
            idx = poly.names.index(diffvar)
        elif isinstance(diffvar, int):
            idx = diffvar
        else:
            diffvar = numpoly.aspolynomial(diffvar)
            assert len(diffvar.names) == 1, "only one at the time"
            assert numpy.all(diffvar.exponents == 1), (
                "derivative variable assumes singletons")
            idx = poly.names.index(diffvar.names[0])

        exponents = poly.exponents
        coefficients = [
            (exponent[idx]*coefficient.T).T
            for exponent, coefficient in zip(exponents, poly.coefficients)
        ]
        exponents[:, idx] -= 1
        assert not numpy.any(exponents < 0)

        poly = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            names=poly_ref.names,
        )
        poly, poly_ref = numpoly.align_polynomials(poly, poly_ref)
    return poly


def gradient(poly):
    """
    Polynomial gradient operator.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to differentiate. If polynomial vector is provided,
            the Jacobi-matrix is returned instead.

    Returns:
        Same as ``poly`` but with an extra first dimensions, one for each
        variable in ``poly.indeterminants``, filled with gradient values.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly = 5*x**5+4
        >>> numpoly.gradient(poly)
        polynomial([25*x**4])
        >>> poly = 4*x**3+2*y**2+3
        >>> numpoly.gradient(poly)
        polynomial([12*x**2, 4*y])
        >>> poly = numpoly.polynomial([1, x**3, x*y**2+1])
        >>> numpoly.gradient(poly)
        polynomial([[0, 3*x**2, y**2],
                    [0, 0, 2*x*y]])

    """
    poly = numpoly.aspolynomial(poly)
    polys = [diff(poly, diffvar)[numpy.newaxis]
             for diffvar in poly.names]
    return numpoly.concatenate(polys, axis=0)


def hessian(poly):
    """
    Construct Hessian matrix of polynomials.

    Make Hessian matrix out of a polynomial. Tensor is returned if polynomial
    is a vector.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to differentiate.

    Returns:
        Same as ``poly`` but with two extra dimensions, one for each
        variable in ``poly.indeterminants``, filled with Hessian values.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> poly = 5*x**5+4
        >>> numpoly.hessian(poly)
        polynomial([[100*x**3]])
        >>> poly = 4*x**3+2*y**2+3
        >>> numpoly.hessian(poly)
        polynomial([[24*x, 0],
                    [0, 4]])
        >>> poly = numpoly.polynomial([1, x, x*y**2+1])
        >>> numpoly.hessian(poly)
        polynomial([[[0, 0, 0],
                     [0, 0, 2*y]],
        <BLANKLINE>
                    [[0, 0, 2*y],
                     [0, 0, 2*x]]])

    """
    return gradient(gradient(poly))



"""Truncation rules for indices."""
import numpy


def cross_truncate(indices, bound, norm):
    r"""
    Truncate of indices using L_p norm.

    .. math:
        L_p(x) = \sum_i |x_i/b_i|^p ^{1/p} \leq 1

    where :math:`b_i` are bounds that each :math:`x_i` should follow.

    Args:
        indices (Sequence[int]):
            Indices to be truncated.
        bound (int, Sequence[int]):
            The bound function for witch the indices can not be larger than.
        norm (float, Sequence[float]):
            The `p` in the `L_p`-norm. Support includes both `L_0` and `L_inf`.

    Returns:
        Boolean indices to ``indices`` with True for each index where the
        truncation criteria holds.

    Examples:
        >>> indices = numpy.array(numpy.mgrid[:10, :10]).reshape(2, -1).T
        >>> indices[cross_truncate(indices, 2, norm=0)].T
        array([[0, 0, 0, 1, 2],
               [0, 1, 2, 0, 0]])
        >>> indices[cross_truncate(indices, 2, norm=1)].T
        array([[0, 0, 0, 1, 1, 2],
               [0, 1, 2, 0, 1, 0]])
        >>> indices[cross_truncate(indices, 2, norm=numpy.inf)].T
        array([[0, 0, 0, 1, 1, 1, 2, 2, 2],
               [0, 1, 2, 0, 1, 2, 0, 1, 2]])

    """
    assert norm >= 0, "negative L_p norm not allowed"
    bound = numpy.asfarray(bound)
    if numpy.any(bound < 0):
        out = numpy.zeros((len(indices),), dtype=bool)
    elif norm == 0:
        out = numpy.sum(indices > 0, axis=-1) <= 1
        out[numpy.any(indices > bound, axis=-1)] = False
    elif norm == numpy.inf:
        out = numpy.max(indices/bound, axis=-1) <= 1
    else:
        out = numpy.sum((indices/bound)**norm, axis=-1)**(1./norm) <= 1
    if not numpy.any(bound < 0):
        out[numpy.all(indices == 0, axis=-1)] = True
    return out



"""Split array into multiple sub-arrays along the 3rd axis (depth)."""
import numpy
import numpoly

from .common import implements


@implements(numpy.dsplit)
def dsplit(ary, indices_or_sections):
    """
    Split array into multiple sub-arrays along the 3rd axis (depth).

    Please refer to the `split` documentation.  `dsplit` is equivalent
    to `split` with ``axis=2``, the array is always split along the third
    axis provided the array dimension is greater than or equal to 3.

    Examples:
        >>> poly = numpoly.monomial(8).reshape(2, 2, 2)
        >>> poly
        polynomial([[[1, q],
                     [q**2, q**3]],
        <BLANKLINE>
                    [[q**4, q**5],
                     [q**6, q**7]]])
        >>> part1, part2 = numpoly.dsplit(poly, 2)
        >>> part1
        polynomial([[[1],
                     [q**2]],
        <BLANKLINE>
                    [[q**4],
                     [q**6]]])
        >>> part2
        polynomial([[[q],
                     [q**3]],
        <BLANKLINE>
                    [[q**5],
                     [q**7]]])

    """
    ary = numpoly.aspolynomial(ary)
    results = numpy.dsplit(ary.values, indices_or_sections=indices_or_sections)
    return [numpoly.aspolynomial(result, names=ary.indeterminants)
            for result in results]



"""Clean up polynomial attributes."""
from six import string_types
import numpy
import numpoly


class PolynomialConstructionError(ValueError):
    """Error related to construction of polynomial."""


def clean_attributes(poly):
    """
    Clean up polynomial attributes.

    Some operations results in polynomial with structures that are redundant.
    This includes extra unused indeterminants, and extra terms consisting of
    only zeros.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to clean up.

    Returns:
        Same as `poly`, but with attributes cleaned up.

    Examples:
        >>> x, _ = numpoly.align_polynomials(*numpoly.symbols("x y"))
        >>> x.indeterminants
        polynomial([x, y])
        >>> x.exponents
        array([[1, 0],
               [0, 1]], dtype=uint32)
        >>> x = numpoly.clean_attributes(x)
        >>> x.indeterminants
        polynomial([x])
        >>> x.exponents
        array([[1]], dtype=uint32)

    """
    return numpoly.ndpoly.from_attributes(
        exponents=poly.exponents,
        coefficients=poly.coefficients,
        names=poly.names,
        dtype=poly.dtype,
        clean=True,
    )


def postprocess_attributes(exponents, coefficients, names=None):
    """
    Clean up polynomial attributes.

    Args:
        exponents (numpy.ndarray):
            The exponents in an integer array with shape ``(N, D)``, where
            ``N`` is the number of terms in the polynomial sum and ``D`` is
            the number of dimensions.
        coefficients (Sequence[numpy.ndarray]):
            The polynomial coefficients. Must correspond to `exponents` by
            having the same length ``N``.
        names (Union[None, Sequence[str], numpoly.ndpoly]):
            The indeterminant names, either as string names or as
            simple polynomials. Must correspond to the exponents by having
            length ``D``.

    Returns:
        (numpoly.ndarray, List[numpy.ndarray], Optional[Tuple[str, ...]]):
            Same as input, but post-processed.

    """
    exponents = numpy.asarray(exponents)
    coefficients = [numpy.asarray(coefficient) for coefficient in coefficients]

    _validate_input(exponents, coefficients)
    if not coefficients:
        assert exponents.shape == (1, 1)
        return exponents, coefficients, names

    exponents, coefficients = list(zip(*[
        (exponent, coefficient)
        for exponent, coefficient in zip(exponents, coefficients)
        if numpy.any(coefficient) or not numpy.any(exponent)
    ]))
    exponents = numpy.asarray(exponents, dtype=int)

    if isinstance(names, numpoly.ndpoly):
        names = names.names
    if isinstance(names, string_types):
        if exponents.shape[1] > 1:
            names = ["%s%d" % (names, idx)
                     for idx in range(exponents.shape[1])]
        else:
            names = [names]

    indices = numpy.any(exponents != 0, 0)
    if not numpy.any(indices):
        indices[0] = True

    if names:
        if len(names) != exponents.shape[1]:
            raise PolynomialConstructionError(
                "Name length incompatible exponent length; "
                "len%s != %d" % (names, exponents.shape[1]))
        names = numpy.array(names)[indices].tolist()
        if sorted(set(names)) != sorted(names):
            raise PolynomialConstructionError(
                "Duplicate indeterminant names: %s" % names)

    exponents = exponents[:, indices]

    exponents_, count = numpy.unique(exponents, return_counts=True, axis=0)
    if numpy.any(count > 1):
        raise PolynomialConstructionError(
            "Duplicate exponent keys found: %s" % exponents_[count > 1][0])

    return exponents, coefficients, names


def _validate_input(exponents, coefficients):
    """Make sure the shape of the input is valid."""
    if exponents.ndim != 2:
        raise PolynomialConstructionError(
            "expected exponents.ndim == 2; found %d" % exponents.ndim)
    if coefficients and len(exponents) != len(coefficients):
        raise PolynomialConstructionError(
            "expected len(exponents) == len(coefficients); found %d != %d" % (
                len(exponents), len(coefficients)))



"""Return true where two arrays are element-wise equal within a tolerance."""
import numpy
import numpoly

from .common import implements, simple_dispatch


@implements(numpy.isclose)
def isclose(a, b, rtol=1e-5, atol=1e-8, equal_nan=False):
    """
    Return true where two arrays are element-wise equal within a tolerance.

    The tolerance values are positive, typically very small numbers. The
    relative difference (`rtol` * abs(`b`)) and the absolute difference `atol`
    are added together to compare against the absolute difference between `a`
    and `b`.

    .. warning:: The default `atol` is not appropriate for comparing numbers
                that are much smaller than one (see Notes).

    Args:
        a, b (numpoly.ndpoly):
            Input arrays to compare.
        rtol (float):
            The relative tolerance parameter (see Notes).
        atol (float):
            The absolute tolerance parameter (see Notes).
        equal_nan (bool):
            Whether to compare NaN's as equal.  If True, NaN's in `a` will be
            considered equal to NaN's in `b` in the output array.

    Returns:
        (numpy.ndarray):
            Returns a boolean array of where `a` and `b` are equal within the
            given tolerance. If both `a` and `b` are scalars, returns a single
            boolean value.

    Notes:
        For finite values, isclose uses the following equation to test whether
        two floating point values are equivalent.

        absolute(`a` - `b`) <= (`atol` + `rtol` * absolute(`b`))

        Unlike the built-in `math.isclose`, the above equation is not symmetric
        in `a` and `b` -- it assumes `b` is the reference value -- so that
        `isclose(a, b)` might be different from `isclose(b, a)`. Furthermore,
        the default value of atol is not zero, and is used to determine what
        small values should be considered close to zero. The default value is
        appropriate for expected values of order unity: if the expected values
        are significantly smaller than one, it can result in false positives.
        `atol` should be carefully selected for the use case at hand. A zero
        value for `atol` will result in `False` if either `a` or `b` is zero.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> numpoly.isclose([1e10*x, 1e-7], [1.00001e10*x, 1e-8])
        array([ True, False])
        >>> numpoly.isclose([1e10*x, 1e-8], [1.00001e10*x, 1e-9])
        array([ True,  True])
        >>> numpoly.isclose([1e10*x, 1e-8], [1.00001e10*y, 1e-9])
        array([False,  True])
        >>> numpoly.isclose([x, numpy.nan],
        ...                 [x, numpy.nan], equal_nan=True)
        array([ True,  True])

    """
    a, b = numpoly.align_polynomials(a, b)
    out = numpy.ones(a.shape, dtype=bool)
    for key in a.keys:
        out &= numpy.isclose(
            a[key], b[key], atol=atol, rtol=rtol, equal_nan=equal_nan)
    return out



"""Implementation wrapper."""
import numpy
import numpoly

ARRAY_FUNCTIONS = {}


def implements(*array_functions):
    """Register an __array_function__ implementation."""
    def decorator(numpoly_function):
        """Register function."""
        for func in array_functions:
            ARRAY_FUNCTIONS[func] = numpoly_function
        return numpoly_function

    return decorator


def simple_dispatch(
        numpy_func,
        inputs,
        out=None,
        **kwargs
):
    """
    Dispatch function between numpy and numpoly.

    Assumes that evaluation can be performed on the coefficients alone and that
    there are no change to the polynomials.

    Args:
        numpy_func (Callable):
            The numpy function to evaluate `inputs` on.
        inputs (Iterable[numpoly.ndpoly]):
            One or more input arrays.
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        kwargs:
            Keyword args passed to `numpy_func`.

    Returns:
        (numpoly.ndpoly):
            Polynomial, where the coefficients from `input` are passed to
            `numpy_func` to create the output coefficients.

    """
    inputs = numpoly.align_polynomials(*inputs)
    no_output = out is None
    for key in inputs[0].keys:

        if out is None:
            tmp = numpy_func(*[poly[key] for poly in inputs], **kwargs)
            out = numpoly.ndpoly(
                exponents=inputs[0].exponents,
                shape=tmp.shape,
                names=inputs[0].indeterminants,
                dtype=tmp.dtype,
            )
            out[key] = tmp
        else:
            tmp = numpy_func(
                *[poly[key] for poly in inputs], out=out[key], **kwargs)

    if no_output:
        out = numpoly.clean_attributes(out)
    return out



"""Return the indices of the elements that are non-zero."""
import numpy
import numpoly

from .common import implements

@implements(numpy.nonzero)
def nonzero(x, **kwargs):
    """
    Return the indices of the elements that are non-zero.

    Args:
        x (numpoly.ndpoly):
            Input array.

    Returns:
        indices (Union[bool, numpoly.ndarray]):
            Indices of elements that are non-zero.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> f = numpoly.polynomial([[3*x, 0, 0], [0, 4*y, 0], [5*x+y, 6*x, 0]])
        >>> f
        polynomial([[3*x, 0, 0],
                    [0, 4*y, 0],
                    [5*x+y, 6*x, 0]])
        >>> numpoly.nonzero(f)
        (array([0, 1, 2, 2]), array([0, 1, 0, 1]))
        >>> f[numpoly.nonzero(f)]
        polynomial([3*x, 4*y, 5*x+y, 6*x])

    """
    a = numpoly.aspolynomial(x)

    return numpy.nonzero(numpy.any(a.coefficients, axis=0))



"""Polynomial functionality."""
from .call import call
from .decompose import decompose
from .derivative import diff, gradient, hessian
from .isconstant import isconstant
from .monomial import monomial, bindex
from .symbols import symbols
from .tonumpy import tonumpy



"""Matrix product of two arrays."""
import numpy
import numpoly

from .common import implements

ERROR_MESSAGE = """\
matmul: Input operand %d does not have enough dimensions \
(has 0, gufunc core with signature (n?,k),(k,m?)->(n?,m?) requires 1)
"""


@implements(numpy.matmul)
def matmul(x1, x2, out=None, **kwargs):
    """
    Matrix product of two arrays.

    Args:
        x1, x2 (numpoly.ndpoly):
            Input arrays, scalars not allowed.
        out (Optional[numpoly.ndpoly]):
            A location into which the result is stored. If provided, it must have
            a shape that matches the signature `(n,k),(k,m)->(n,m)`. If not
            provided or `None`, a freshly-allocated array is returned.
            n,k,m

    Returns:
        (numpoly.ndpoly):
            The matrix product of the inputs. This is a scalar only when both
            x1, x2 are 1-d vectors.

    Raises:
        ValueError:
            If the last dimension of `x1` is not the same size as
            the second-to-last dimension of `x2`.

    Examples:
        >>> poly = numpoly.symbols("q:4").reshape(2, 2)
        >>> poly
        polynomial([[q0, q1],
                    [q2, q3]])
        >>> numpoly.matmul(poly, [[0, 1], [2, 3]])
        polynomial([[2*q1, 3*q1+q0],
                    [2*q3, 3*q3+q2]])
        >>> numpoly.matmul(poly, [4, 5])
        polynomial([[4*q1+4*q0, 5*q1+5*q0],
                    [4*q3+4*q2, 5*q3+5*q2]])
        >>> numpoly.matmul(*poly)
        polynomial([q1*q2+q0*q2, q1*q3+q0*q3])

    """
    x1 = numpoly.aspolynomial(x1)
    x2 = numpoly.aspolynomial(x2)
    if not x1.shape:
        raise ValueError(ERROR_MESSAGE % 0)
    if not x2.shape:
        raise ValueError(ERROR_MESSAGE % 1)
    x1 = numpoly.reshape(x1, x1.shape+(1,))
    x2 = numpoly.reshape(x2, x2.shape[:-2]+(1,)+x2.shape[-2:])
    x1, x2 = numpoly.broadcast_arrays(x1, x2)
    out = numpoly.multiply(x1, x2, out=out, **kwargs)
    out = numpoly.sum(out, axis=-2)
    return out



"""Construct symbol variables."""
import re
import string
from six import string_types

import numpy
import numpoly


def symbols(names=None, asarray=False, dtype="i8"):
    """
    Construct symbol variables.

    Most directly be providing a list of string names. But a set of shorthands
    also exists:

    * ``,`` and `` `` (space) can be used as a variable delimiter.
    * ``{number}:{number}`` can be used to define a numerical range.
    * ``{letter}:{letter}`` can be used to define a alphabet range.

    Args:
        names (None, str, Tuple[str, ...]):
            Indeterminants are determined by splitting the string on space. If
            iterable of strings, indeterminants defined directly.
        asarray (bool):
            Enforce output as array even in the case where there is only one
            variable.
        dtype (numpy.dtype):
            The data type of the polynomial coefficients.

    Returns:
        (numpoly.ndpoly):
            Polynomial array with unit components in each dimension.

    Examples:
        >>> numpoly.symbols()
        polynomial(q)
        >>> numpoly.symbols("z,")
        polynomial([z])
        >>> numpoly.symbols("z", asarray=True)
        polynomial([z])
        >>> numpoly.symbols(["alpha", "beta"])
        polynomial([alpha, beta])
        >>> numpoly.symbols("x y z")
        polynomial([x, y, z])
        >>> numpoly.symbols("a,b,c")
        polynomial([a, b, c])
        >>> numpoly.symbols("q:7")
        polynomial([q0, q1, q2, q3, q4, q5, q6])
        >>> numpoly.symbols("q3:6")
        polynomial([q3, q4, q5])
        >>> numpoly.symbols("za:f")
        polynomial([za, zb, zc, zd, ze, zf])

    """
    if names is None:
        coefficients = numpy.ones((1, 1) if asarray else 1, dtype=dtype)
        return numpoly.ndpoly.from_attributes(
            exponents=[(1,)], coefficients=coefficients)

    if not isinstance(names, string_types):
        names = list(names)

    else:
        names = re.sub(" ", ",", names)
        if "," in names:
            asarray = True
            names = [name for name in names.split(",") if name]

        elif re.search(r"(\d*):(\d+)", names):

            match = re.search(r"(\d*):(\d+)", names)
            start = int(match.group(1) or 0)
            end = int(match.group(2))
            names = [
                names.replace(match.group(0), str(idx))
                for idx in range(start, end)
            ]

        elif re.search(r"([a-zA-Z]+):([a-zA-Z]+)", names):

            match = re.search(r"([a-zA-Z]):([a-zA-Z])", names)
            start = string.ascii_letters.index(match.group(1))
            end = string.ascii_letters.index(match.group(2))
            names = [
                names.replace(
                    match.group(0), string.ascii_letters[idy])
                for idy in range(start, end+1)
            ]

    exponents = numpy.eye(len(names), dtype=int)
    coefficients = numpy.eye(len(names), dtype=dtype)
    if len(names) == 1 and not asarray:
        coefficients = coefficients[0]
    return numpoly.ndpoly.from_attributes(
        exponents=exponents,
        coefficients=coefficients,
        names=names,
    )



"""Compute the outer product of two vectors."""
import numpy
import numpoly

from .common import implements


@implements(numpy.outer)
def outer(a, b, out=None):
    """
    Compute the outer product of two vectors.

    Given two vectors, ``a = [a0, a1, ..., aM]`` and
    ``b = [b0, b1, ..., bN]``, the outer product is::

        [[a0*b0  a0*b1 ... a0*bN ]
         [a1*b0    .
         [ ...          .
         [aM*b0            aM*bN ]]

    Args:
        a (numpoly.ndpoly):
            First input vector. Input is flattened if not already
            1-dimensional.
        b (numpoly.ndpoly):
            Second input vector. Input is flattened if not already
            1-dimensional.
        out (numpy.ndarray):
            A location where the result is stored.

    Returns:
        (numpoly.ndpoly):
            ``out[i, j] = a[i] * b[j]``

    Examples:
        >>> numpoly.outer(numpoly.symbols("x y z"), numpy.arange(5))
        polynomial([[0, x, 2*x, 3*x, 4*x],
                    [0, y, 2*y, 3*y, 4*y],
                    [0, z, 2*z, 3*z, 4*z]])

    """
    a, b = numpoly.align_exponents(a, b)
    a = a.flatten()[:, numpy.newaxis]
    b = b.flatten()[numpy.newaxis, :]
    return numpoly.multiply(a, b, out=out)



"""Broadcast any number of arrays against each other."""
import numpy
import numpoly

from .common import implements


@implements(numpy.broadcast_arrays)
def broadcast_arrays(*args, **kwargs):
    """
    Broadcast any number of arrays against each other.

    Args:
        args (numpoly.ndpoly):
            The arrays to broadcast.
        subok (bool):
            If True, then sub-classes will be passed-through, otherwise the
            returned arrays will be forced to be a base-class array (default).

    Returns:
        (List[numpoly.ndpoly]):
            These arrays are views on the original arrays.  They are typically
            not contiguous.  Furthermore, more than one element of a
            broadcasted array may refer to a single memory location. If you
            need to write to the arrays, make copies first. While you can set
            the ``writable`` flag True, writing to a single output value may
            end up changing more than one location in the output array.

    Examples:
        >>> poly1 = numpoly.monomial(3).reshape(1, 3)
        >>> poly1
        polynomial([[1, q, q**2]])
        >>> poly2 = numpoly.monomial(2).reshape(2, 1)
        >>> poly2
        polynomial([[1],
                    [q]])
        >>> result1, result2 = numpoly.broadcast_arrays(poly1, poly2)
        >>> result1
        polynomial([[1, q, q**2],
                    [1, q, q**2]])
        >>> result2
        polynomial([[1, 1, 1],
                    [q, q, q]])

    """
    args = [numpoly.aspolynomial(arg) for arg in args]
    results = numpy.broadcast_arrays(*[arg.values for arg in args], **kwargs)
    return [numpoly.aspolynomial(result, names=arg.indeterminants)
            for result, arg in zip(results, args)]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['numpoly',
 'numpoly.array_function',
 'numpoly.construct',
 'numpoly.poly_function',
 'numpoly.poly_function.monomial']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.16,<2.0', 'six']

setup_kwargs = {
    'name': 'numpoly',
    'version': '0.1.14',
    'description': 'Polynomials as a numpy datatype',
    'long_description': '.. image:: doc/.static/numpoly_logo.svg\n   :height: 300 px\n   :width: 300 px\n   :align: center\n\n|circleci| |codecov| |pypi| |readthedocs|\n\n.. |circleci| image:: https://circleci.com/gh/jonathf/numpoly/tree/master.svg?style=shield\n    :target: https://circleci.com/gh/jonathf/numpoly/tree/master\n.. |codecov| image:: https://codecov.io/gh/jonathf/numpoly/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/jonathf/numpoly\n.. |pypi| image:: https://badge.fury.io/py/numpoly.svg\n    :target: https://badge.fury.io/py/numpoly\n.. |readthedocs| image:: https://readthedocs.org/projects/numpoly/badge/?version=master\n    :target: http://numpoly.readthedocs.io/en/master/?badge=master\n\nNumpoly is a generic library for creating, manipulating and evaluating\narrays of polynomials.\n\nThe polynomial base class ``numpoly.ndpoly`` is a subclass of ``numpy.ndarray``\nimplemented to represent polynomials as array element. This makes the library\nvery fast with the respect of the size of the coefficients. It is also adds\ncompatibility with ``numpy`` functions and methods, where that makes sense,\nmaking the interface more intuitive.\n\nMany numerical analysis, polynomial approximations as proxy predictors for real\npredictors to do analysis on. These models are often solutions to non-linear\nproblems discretized with high mesh. As such, the corresponding polynomial\napproximation consist of high number of dimensions and large multi-dimensional\npolynomial coefficients. For these kind of problems ``numpoly`` is a good fit.\n\nOne example where ``numpoly`` is used as the backend is the uncertainty\nquantification library `chaospy <https://github.com/jonathf/chaospy>`_.\n\n.. contents:: Table of Contents:\n\nInstallation\n------------\n\nInstallation should be straight forward:\n\n.. code-block:: bash\n\n    pip install numpoly\n\nAnd you should be ready to go.\n\nExample usage\n-------------\n\nConstructing polynomial is typically done using one of the available\nconstructors:\n\n.. code-block:: python\n\n   >>> numpoly.monomial(start=0, stop=4, names=("x", "y"))\n   polynomial([1, y, x, y**2, x*y, x**2, y**3, x*y**2, x**2*y, x**3])\n\nIt is also possible to construct your own from symbols:\n\n.. code-block:: python\n\n   >>> x, y = numpoly.symbols("x y")\n   >>> numpoly.polynomial([1, x**2-1, x*y, y**2-1])\n   polynomial([1, -1+x**2, x*y, -1+y**2])\n\nOr in combination with numpy objects using various arithmetics:\n\n.. code-block:: python\n\n   >>> x**numpy.arange(4)-y**numpy.arange(3, -1, -1)\n   polynomial([1-y**3, x-y**2, x**2-y, -1+x**3])\n\nThe constructed polynomials can be evaluated as needed:\n\n.. code-block:: python\n\n   >>> poly = 3*x+2*y+1\n   >>> poly(x=y, y=[1, 2, 3])\n   polynomial([3+3*y, 5+3*y, 7+3*y])\n\nOr manipulated using various numpy functions:\n\n.. code-block:: python\n\n   >>> numpy.reshape(x**numpy.arange(4), (2, 2))\n   polynomial([[1, x],\n               [x**2, x**3]])\n   >>> numpy.sum(numpoly.monomial(13, names="z")[::3])\n   polynomial(1+z**3+z**6+z**9+z**12)\n\nIn addition there are also several operators specific to the polynomial:\n\n.. code-block:: python\n\n   >>> numpoly.diff([1, x, x**2], x)\n   polynomial([0, 1, 2*x])\n   >>> numpoly.gradient([x*y, x+y])\n   polynomial([[y, 1],\n               [x, 1]])\n\nDevelopment\n-----------\n\nDevelopment is done using `Poetry <https://poetry.eustace.io/>`_ manager.\nInside the repository directory, install and create a virtual environment with:\n\n.. code-block:: bash\n\n   poetry install\n\nTo run tests, run:\n\n.. code-block:: bash\n\n   poetry run pytest numpoly test doc --doctest-modules\n\nQuestions & Troubleshooting\n---------------------------\n\nFor any problems and questions you might have related to ``numpoly``, please\nfeel free to file an `issue <https://github.com/jonathf/numpoly/issues>`_.\n',
    'author': 'Jonathan Feinberg',
    'author_email': 'jonathf@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jonathf/numpoly',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
}


setup(**setup_kwargs)



"""Stack arrays in sequence vertically (row wise)."""
import numpy
import numpoly

from .common import implements


@implements(numpy.vstack)
def vstack(tup):
    """
    Stack arrays in sequence vertically (row wise).

    This is equivalent to concatenation along the first axis after 1-D arrays
    of shape `(N,)` have been reshaped to `(1,N)`. Rebuilds arrays divided by
    `vsplit`.

    This function makes most sense for arrays with up to 3 dimensions. For
    instance, for pixel-data with a height (first axis), width (second axis),
    and r/g/b channels (third axis). The functions `concatenate`, `stack` and
    `block` provide more general stacking and concatenation operations.

    Args:
        tup (Sequence[numpoly.ndpoly]):
            The arrays must have the same shape along all but the first axis.
            1-D arrays must have the same length.

    Returns:
        (numpoly.ndpoly):
            The array formed by stacking the given arrays, will be at least
            2-D.

    Examples:
        >>> a = numpoly.symbols("x y z")
        >>> b = numpoly.polynomial([1, 2, 3])
        >>> numpoly.vstack([a, b])
        polynomial([[x, y, z],
                    [1, 2, 3]])
        >>> c = numpoly.polynomial([[1], [2], [3]])
        >>> d = a.reshape(3, 1)
        >>> numpoly.vstack([c, d])
        polynomial([[1],
                    [2],
                    [3],
                    [x],
                    [y],
                    [z]])

    """
    arrays = numpoly.align_exponents(*tup)
    arrays = numpoly.align_dtype(*arrays)
    result = numpy.vstack([array.values for array in arrays])
    return numpoly.aspolynomial(result, names=arrays[0].indeterminants)



"""Split an array into multiple sub-arrays."""
import numpy
import numpoly

from .common import implements


@implements(numpy.split)
def split(ary, indices_or_sections, axis=0):
    """
    Split an array into multiple sub-arrays.

    Args:
        ary (numpoly.ndpoly):
            Array to be divided into sub-arrays.
        indices_or_sections (int, Sequence[int]):
            If `indices_or_sections` is an integer, N, the array will be
            divided into N equal arrays along `axis`.  If such a split is not
            possible, an error is raised.

            If `indices_or_sections` is a 1-D array of sorted integers, the
            entries indicate where along `axis` the array is split.  For
            example, ``[2, 3]`` would, for ``axis=0``, result in

            - ary[:2]
            - ary[2:3]
            - ary[3:]

            If an index exceeds the dimension of the array along `axis`, an
            empty sub-array is returned correspondingly.
        axis (int):
            The axis along which to split, default is 0.

    Returns:
        (List[numpoly.ndpoly])
            A list of sub-arrays.

    Raises:
        ValueError:
            If `indices_or_sections` is given as an integer, but a split does
            not result in equal division.

    See Also:
        array_split:
            Split an array into multiple sub-arrays of equal or near-equal
            size. Does not raise an exception if an equal division cannot be
            made.
        hsplit:
            Split array into multiple sub-arrays horizontally (column-wise).
        vsplit:
            Split array into multiple sub-arrays vertically (row wise).
        dsplit:
            Split array into multiple sub-arrays along the 3rd axis (depth).

    Examples:
        >>> poly = numpoly.monomial(16).reshape(4, 4)
        >>> poly
        polynomial([[1, q, q**2, q**3],
                    [q**4, q**5, q**6, q**7],
                    [q**8, q**9, q**10, q**11],
                    [q**12, q**13, q**14, q**15]])
        >>> part1, _ = numpoly.split(poly, 2, axis=0)
        >>> part1
        polynomial([[1, q, q**2, q**3],
                    [q**4, q**5, q**6, q**7]])
        >>> part1, _ = numpoly.split(poly, 2, axis=1)
        >>> part1
        polynomial([[1, q],
                    [q**4, q**5],
                    [q**8, q**9],
                    [q**12, q**13]])

    """
    ary = numpoly.aspolynomial(ary)
    results = numpy.split(
        ary.values, indices_or_sections=indices_or_sections, axis=axis)
    return [numpoly.aspolynomial(result, names=ary.indeterminants)
            for result in results]



"""Compute the truth value of x1 AND x2 element-wise."""
import numpy
import numpoly

from .common import implements, simple_dispatch


@implements(numpy.logical_and)
def logical_and(x1, x2, out=None, where=True, **kwargs):
    """
    Compute the truth value of x1 AND x2 element-wise.

    Args:
        x1, x2 (numpoly.ndpoly):
            Input arrays. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        y (numpy.ndarray):
            Boolean result of the logical OR operation applied to the elements
            of `x1` and `x2`; the shape is determined by broadcasting. This is
            a scalar if both `x1` and `x2` are scalars.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> numpoly.logical_and(x, 0)
        False
        >>> numpoly.logical_and([x, False], [x, y])
        array([ True, False])
        >>> z = numpy.arange(5)
        >>> numpoly.logical_and(z > 1, z < 4)
        array([False, False,  True,  True, False])

    """
    x1 = numpoly.aspolynomial(x1)
    x2 = numpoly.aspolynomial(x2)
    coefficients1 = numpy.any(x1.coefficients, 0)
    coefficients2 = numpy.any(x2.coefficients, 0)
    return numpy.logical_and(
        coefficients1, coefficients2, out=out, where=where, **kwargs)



"""Decompose a polynomial to component form."""
import numpy
import numpoly


def decompose(poly):
    """
    Decompose a polynomial to component form.

    In array missing values are padded with 0 to make decomposition compatible
    with ``chaospy.sum(ouput, 0)``.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to decompose

    Returns:
        (numpoly.ndpoly):
            Decomposed polynomial with ``poly.shape==(M,)+output.shape``
            where ``M`` is the number of components in `poly`.

    Examples:
        >>> x = numpoly.symbols()
        >>> poly = numpoly.polynomial([x**2-1, 2])
        >>> poly
        polynomial([-1+q**2, 2])
        >>> numpoly.decompose(poly)
        polynomial([[-1, 2],
                    [q**2, 0]])
        >>> numpoly.sum(numpoly.decompose(poly), 0)
        polynomial([-1+q**2, 2])

    """
    return numpoly.concatenate([
        numpoly.construct.polynomial_from_attributes(
            exponents=[expon],
            coefficients=[poly[key]],
            names=poly.indeterminants,
            clean=False,
        )[numpy.newaxis] for key, expon in zip(poly.keys, poly.exponents)
    ])



"""Collection of numpy wrapper functions."""
from .common import ARRAY_FUNCTIONS

from .absolute import absolute as abs, absolute
from .add import add
from .any import any
from .all import all
from .allclose import allclose
from .around import around as round, around
from .array_repr import array_repr
from .array_split import array_split
from .array_str import array_str
from .atleast_1d import atleast_1d
from .atleast_2d import atleast_2d
from .atleast_3d import atleast_3d
from .broadcast_arrays import broadcast_arrays
from .ceil import ceil
from .choose import choose
from .common_type import common_type
from .concatenate import concatenate
from .count_nonzero import count_nonzero
from .cumsum import cumsum
from .divide import divide
from .dsplit import dsplit
from .dstack import dstack
from .equal import equal
from .floor import floor
from .floor_divide import floor_divide
from .hstack import hstack
from .hsplit import hsplit
from .inner import inner
from .isclose import isclose
from .isfinite import isfinite
from .logical_and import logical_and
from .logical_or import logical_or
from .matmul import matmul
from .mean import mean
from .moveaxis import moveaxis
from .multiply import multiply
from .negative import negative
from .nonzero import nonzero
from .not_equal import not_equal
from .outer import outer
from .positive import positive
from .power import power
from .prod import prod
from .repeat import repeat
from .reshape import reshape
from .rint import rint
from .split import split
from .square import square
from .stack import stack
from .subtract import subtract
from .sum import sum
from .tile import tile
from .transpose import transpose
from .vsplit import vsplit
from .vstack import vstack



"""Counts the number of non-zero values in the array a."""
import numpy
import numpoly

from .common import implements

@implements(numpy.count_nonzero)
def count_nonzero(x, axis=None, **kwargs):
    """
    Count the number of non-zero values in the array a.

    Args:
        x (numpoly.ndpoly):
            The array for which to count non-zeros.
        axis: (Union[int, Tuple[int], None]):
            Axis or tuple of axes along which to count non-zeros. Default is
            None, meaning that non-zeros will be counted along a flattened
            version of a.

    Returns:
        count (Union[bool, numpy.ndarray]):
            Number of non-zero values in the array along a given axis.
            Otherwise, the total number of non-zero values in the array is
            returned.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> numpoly.count_nonzero([x])
        1
        >>> numpoly.count_nonzero([[0,x,x*x,0,0],[x+1,0,0,2*x,19*x]])
        5
        >>> numpoly.count_nonzero([[0,x,7*x,0,0],[3*y,0,0,2,19*x+y]], axis=0)
        array([1, 1, 1, 1, 1])
        >>> numpoly.count_nonzero([[0,x,y,0,0],[x,0,0,2*x,19*y]], axis=1)
        array([2, 3])

    """
    a = numpoly.aspolynomial(x)

    return numpy.count_nonzero(numpy.any(a.coefficients, axis=0), axis=axis)



"""Stack arrays in sequence horizontally (column wise)."""
import numpy
import numpoly

from .common import implements


@implements(numpy.hstack)
def hstack(tup):
    """
    Stack arrays in sequence horizontally (column wise).

    This is equivalent to concatenation along the second axis, except for 1-D
    arrays where it concatenates along the first axis. Rebuilds arrays divided
    by `hsplit`.

    This function makes most sense for arrays with up to 3 dimensions. For
    instance, for pixel-data with a height (first axis), width (second axis),
    and r/g/b channels (third axis). The functions `concatenate`, `stack` and
    `block` provide more general stacking and concatenation operations.

    Args:
        tup (Sequence[numpoly.ndpoly]):
            The arrays must have the same shape along all but the second axis,
            except 1-D arrays which can be any length.

    Returns:
        (numpoly.ndpoly):
            The array formed by stacking the given arrays.

    Examples:
        >>> a = numpoly.symbols("x y z")
        >>> b = numpoly.polynomial([1, 2, 3])
        >>> numpoly.hstack([a, b])
        polynomial([x, y, z, 1, 2, 3])
        >>> c = numpoly.polynomial([[1], [2], [3]])
        >>> d = a.reshape(3, 1)
        >>> numpoly.hstack([c, d])
        polynomial([[1, x],
                    [2, y],
                    [3, z]])

    """
    arrays = numpoly.align_exponents(*tup)
    arrays = numpoly.align_dtype(*arrays)
    result = numpy.hstack([array.values for array in arrays])
    return numpoly.aspolynomial(
        result, names=arrays[0].indeterminants)



"""Test element-wise for finiteness (not infinity or not Not a Number)."""
import numpy
import numpoly

from .common import implements, simple_dispatch


@implements(numpy.isfinite)
def isfinite(x, out=None, where=True, **kwargs):
    """
    Test element-wise for finiteness (not infinity or not Not a Number).

    The result is returned as a boolean array.

    Args:
        x (numpoly.ndpoly):
            Input values.
        out (Optional[numpoly.ndpoly]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value.
            Note that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (numpy.ndarray):
            True where ``x`` is not positive infinity, negative infinity, or
            NaN; false otherwise. This is a scalar if `x` is a scalar.

    Notes:
        Not a Number, positive infinity and negative infinity are considered to
        be non-finite.

    Examples:
        >>> numpoly.isfinite(1)
        True
        >>> numpoly.isfinite(0)
        True
        >>> numpoly.isfinite(numpy.nan*numpoly.symbols("x"))
        False
        >>> numpoly.isfinite(numpy.inf)
        False
        >>> numpoly.isfinite(numpy.NINF)
        False
        >>> numpoly.isfinite([numpy.log(-1.), 1., numpy.log(0)])
        array([False,  True, False])

    """
    out = simple_dispatch(
        numpy_func=numpy.isfinite,
        inputs=(x,),
        out=out,
        where=where,
        **kwargs
    )
    return numpy.all(out.coefficients, axis=0)



"""Return a true division of the inputs, element-wise."""
from __future__ import division

import numpy
import numpoly

from .common import implements


@implements(numpy.divide, numpy.true_divide)
def divide(x1, x2, out=None, where=True, **kwargs):
    """
    Return a true division of the inputs, element-wise.

    Instead of the Python traditional 'floor division', this returns a true
    division.  True division adjusts the output type to present the best
    answer, regardless of input types.

    Args:
        x1 (numpoly.ndpoly):
            Dividend array.
        x2 (numpoly.ndpoly):
            Divisor array. If ``x1.shape != x2.shape``, they must be
            broadcastable to a commo n shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (numpoly.ndpoly):
            This is a scalar if both `x1` and `x2` are scalars.

    Examples:
        >>> xyz = numpoly.symbols("x y z")
        >>> numpoly.divide(xyz, 4)
        polynomial([0.25*x, 0.25*y, 0.25*z])
        >>> numpoly.divide(xyz, [1, 2, 4])
        polynomial([x, 0.5*y, 0.25*z])
        >>> numpoly.divide([1, 2, 4], xyz)
        Traceback (most recent call last):
            ...
        ValueError: only constant polynomials can be converted to array.

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    x2 = x2.tonumpy()
    no_output = out is None
    if no_output:
        out = numpoly.ndpoly(
            exponents=x1.exponents,
            shape=x1.shape,
            names=x1.indeterminants,
            dtype=numpy.common_type(x1, numpy.array(1.)),
        )
    elif not isinstance(out, numpy.ndarray):
        assert len(out) == 1, "only one output"
        out = out[0]
    for key in x1.keys:
        out[key] = 0
        numpy.true_divide(x1[key], x2, out=out[key], where=where, **kwargs)
    if no_output:
        out = numpoly.clean_attributes(out)
    return out



"""Return (x1 != x2) element-wise."""
import numpy
import numpoly

from .common import implements


@implements(numpy.not_equal)
def not_equal(x1, x2, out=None, where=True, **kwargs):
    """
    Return (x1 != x2) element-wise.

    Args:
        x1, x2 (numpoly.ndpoly):
            Input arrays.  If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (numpy.ndarray):
            Output array, element-wise comparison of `x1` and `x2`. Typically
            of type bool, unless ``dtype=object`` is passed.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> numpoly.not_equal([x, x], [x, y])
        array([False,  True])
        >>> numpoly.not_equal([x, x], [[x, y], [y, x]])
        array([[False,  True],
               [ True, False]])

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    for key in x1.keys:
        tmp = numpy.not_equal(x1[key], x2[key], where=where, **kwargs)
        if out is None:
            out = tmp
        else:
            out |= tmp
    return out
